@@ -1,19 +1,20 @@
 from __future__ import unicode_literals
 
 from prompt_toolkit.enums import DEFAULT_BUFFER
-from prompt_toolkit.filters import IsDone, HasCompletions, RendererHeightIsKnown, Always, HasFocus
+from prompt_toolkit.filters import IsDone, HasCompletions, RendererHeightIsKnown, Always, HasFocus, Condition
 from prompt_toolkit.key_binding.vi_state import InputMode
 from prompt_toolkit.layout import Window, HSplit, VSplit, FloatContainer, Float
 from prompt_toolkit.layout.controls import BufferControl, TokenListControl, FillControl
 from prompt_toolkit.layout.dimension import LayoutDimension
-from prompt_toolkit.layout.menus import CompletionsMenu
+from prompt_toolkit.layout.menus import CompletionsMenu, MultiColumnCompletionsMenu
 from prompt_toolkit.layout.processors import HighlightSearchProcessor, HighlightSelectionProcessor, HighlightMatchingBracketProcessor, ConditionalProcessor
 from prompt_toolkit.layout.screen import Char
 from prompt_toolkit.layout.toolbars import CompletionsToolbar, ArgToolbar, SearchToolbar, ValidationToolbar, SystemToolbar, TokenListToolbar
 from prompt_toolkit.layout.utils import token_list_width
+from prompt_toolkit.reactive import Integer
 from prompt_toolkit.selection import SelectionType
 
-from ptpython.filters import HasSignature, ShowCompletionsMenu, ShowCompletionsToolbar, ShowSidebar, ShowLineNumbersFilter, ShowSignature, ShowDocstring
+from ptpython.filters import HasSignature, ShowSidebar, ShowLineNumbersFilter, ShowSignature, ShowDocstring
 
 from pygments.lexers import PythonLexer
 from pygments.token import Token
@@ -23,65 +24,152 @@ import sys
 
 __all__ = (
     'create_layout',
+    'CompletionVisualisation',
 )
 
 
-class PythonSidebarControl(TokenListControl):
-    def __init__(self, settings, key_bindings_manager):
-        def get_tokens(cli):
-            tokens = []
-            TB = Token.Sidebar
+class CompletionVisualisation:
+    " Visualisation method for the completions. "
+    NONE = 'none'
+    POP_UP = 'pop-up'
+    MULTI_COLUMN = 'multi-column'
+    TOOLBAR = 'toolbar'
 
-            if key_bindings_manager.enable_vi_mode:
-                mode = 'vi'
-            else:
-                mode = 'emacs'
 
-            if settings.show_completions_toolbar:
-                completion_style = 'toolbar'
-            elif settings.show_completions_menu:
-                completion_style = 'pop-up'
-            else:
-                completion_style = 'off'
+def show_completions_toolbar(python_input):
+    return Condition(lambda cli: python_input.completion_visualisation == CompletionVisualisation.TOOLBAR)
 
-            def append(shortcut, label, status):
-                tokens.append((TB.Shortcut, ' [%s] ' % shortcut))
-                tokens.append((TB.Label, '%-21s' % label))
-                if status:
-                    tokens.append((TB.Status, '%9s\n' % status))
-                else:
-                    tokens.append((TB.Label, '\n'))
 
-            append('F3', 'Completion menu', '(%s)' % completion_style)
-            append('F4', 'Input mode', '(%s)' % mode)
-            append('F5', 'Complete while typing', '(on)' if settings.complete_while_typing else '(off)')
-            append('F6', 'Paste mode', '(on)' if settings.paste_mode else '(off)')
-            append('F8', 'Show signature', '(on)' if settings.show_signature else '(off)')
-            append('F9', 'Show docstring', '(on)' if settings.show_docstring else '(off)')
-            append('F10', 'Show line numbers', '(on)' if settings.show_line_numbers else '(off)')
+def show_completions_menu(python_input):
+    return Condition(lambda cli: python_input.completion_visualisation == CompletionVisualisation.POP_UP)
 
-            return tokens
 
-        super(PythonSidebarControl, self).__init__(get_tokens, Char(token=Token.Sidebar))
+def show_multi_column_completions_menu(python_input):
+    return Condition(lambda cli: python_input.completion_visualisation == CompletionVisualisation.MULTI_COLUMN)
 
 
 class PythonSidebar(Window):
-    def __init__(self, settings, key_bindings_manager):
+    """
+    Sidebar containing the configurable options.
+    """
+    def __init__(self, python_input):
+        def get_tokens(cli):
+            tokens = []
+            T = Token.Sidebar
+
+            def append_category(category):
+                tokens.extend([
+                    (T, '  '),
+                    (T.Title, '   %-36s' % category.title),
+                    (T, '\n'),
+                ])
+
+            def append(selected, label, status):
+                token = T.Selected if selected else T
+
+                tokens.append((T, ' >' if selected else '  '))
+                tokens.append((token.Label, '%-24s' % label))
+                tokens.append((token.Status, ' '))
+                tokens.append((token.Status, '%s' % status))
+
+                if selected:
+                    tokens.append((Token.SetCursorPosition, ''))
+
+                tokens.append((token.Status, ' ' * (14 - len(status))))
+                tokens.append((T, '<' if selected else ''))
+                tokens.append((T, '\n'))
+
+            i = 0
+            for category in python_input.options:
+                append_category(category)
+
+                for option in category.options:
+                    append(i == python_input.selected_option_index,
+                           option.title, '%s' % option.get_current_value())
+                    i += 1
+
+            tokens.pop()  # Remove last newline.
+
+            return tokens
+
         super(PythonSidebar, self).__init__(
-            PythonSidebarControl(settings, key_bindings_manager),
-            width=LayoutDimension.exact(37),
-            filter=ShowSidebar(settings) & ~IsDone())
+            TokenListControl(get_tokens, Char(token=Token.Sidebar),
+                has_focus=ShowSidebar(python_input) & ~IsDone()),
+            width=LayoutDimension.exact(43),
+            height=LayoutDimension(min=3),
+            scroll_offset=1,
+            filter=ShowSidebar(python_input) & ~IsDone())
 
 
-class SignatureControl(TokenListControl):
-    def __init__(self, settings):
+class PythonSidebarNavigation(Window):
+    """
+    Showing the navigation information for the sidebar.
+    """
+    def __init__(self, python_input):
+        def get_tokens(cli):
+            tokens = []
+            T = Token.Sidebar
+
+            # Show navigation info.
+            tokens.extend([
+                (T.Separator , ' ' * 43 + '\n'),
+                (T, '    '),
+                (T.Key, '[Arrows]'),
+                (T, ' '),
+                (T.Key.Description, 'Navigate'),
+                (T, ' '),
+                (T.Key, '[Enter]'),
+                (T, ' '),
+                (T.Key.Description, 'Hide menu'),
+            ])
+
+            return tokens
+
+        super(PythonSidebarNavigation, self).__init__(
+            TokenListControl(get_tokens, Char(token=Token.Sidebar)),
+            width=LayoutDimension.exact(43),
+            height=LayoutDimension.exact(2),
+            filter=ShowSidebar(python_input) & ~IsDone())
+
+
+class PythonSidebarHelp(Window):
+    """
+    Help text for the current item in the sidebar.
+    """
+    def __init__(self, python_input):
+        token = Token.Sidebar.HelpText
+
+        def get_current_description():
+            """
+            Return the description of the selected option.
+            """
+            i = 0
+            for category in python_input.options:
+                for option in category.options:
+                    if i == python_input.selected_option_index:
+                        return option.description
+                    i += 1
+            return ''
+
+        def get_tokens(cli):
+            return [(token, get_current_description())]
+
+        super(PythonSidebarHelp, self).__init__(
+            TokenListControl(get_tokens, Char(token=token)),
+            height=LayoutDimension(min=3),
+            filter=ShowSidebar(python_input) &
+                   Condition(lambda cli: python_input.show_sidebar_help) & ~IsDone())
+
+
+class SignatureToolbar(Window):
+    def __init__(self, python_input):
         def get_tokens(cli):
             result = []
             append = result.append
             Signature = Token.Toolbar.Signature
 
-            if settings.signatures:
-                sig = settings.signatures[0]  # Always take the first one.
+            if python_input.signatures:
+                sig = python_input.signatures[0]  # Always take the first one.
 
                 append((Signature, ' '))
                 try:
@@ -94,12 +182,19 @@ class SignatureControl(TokenListControl):
                 append((Signature.Operator, '('))
 
                 for i, p in enumerate(sig.params):
-                    if i == sig.index:
+                    # Workaround for #47: 'p' is None when we hit the '*' in the signature.
+                    #                     and sig has no 'index' attribute.
+                    # See: https://github.com/jonathanslenders/ptpython/issues/47
+                    #      https://github.com/davidhalter/jedi/issues/598
+                    description = (p.description if p else '*') #or '*'
+                    sig_index = getattr(sig, 'index', 0)
+
+                    if i == sig_index:
                         # Note: we use `_Param.description` instead of
                         #       `_Param.name`, that way we also get the '*' before args.
-                        append((Signature.CurrentName, str(p.description)))
+                        append((Signature.CurrentName, str(description)))
                     else:
-                        append((Signature, str(p.description)))
+                        append((Signature, str(description)))
                     append((Signature.Operator, ', '))
 
                 if sig.params:
@@ -110,21 +205,17 @@ class SignatureControl(TokenListControl):
                 append((Signature, ' '))
             return result
 
-        super(SignatureControl, self).__init__(get_tokens)
-
-
-class SignatureToolbar(Window):
-    def __init__(self, settings):
         super(SignatureToolbar, self).__init__(
-            SignatureControl(settings),
+            TokenListControl(get_tokens),
             height=LayoutDimension.exact(1),
             filter=
                 # Show only when there is a signature
-                HasSignature(settings) &
+                HasSignature(python_input) &
                 # And there are no completions to be shown. (would cover signature pop-up.)
-                (~HasCompletions() | ~ShowCompletionsMenu(settings))
+                ~(HasCompletions() & (show_completions_menu(python_input) |
+                                       show_multi_column_completions_menu(python_input)))
                 # Signature needs to be shown.
-                & ShowSignature(settings) &
+                & ShowSignature(python_input) &
                 # Not done yet.
                 ~IsDone())
 
@@ -133,15 +224,19 @@ class PythonPrompt(TokenListControl):
     """
     Prompt showing something like "In [1]:".
     """
-    def __init__(self, settings):
+    def __init__(self, python_input):
         def get_tokens(cli):
-            return [(Token.Layout.Prompt, 'In [%s]: ' % settings.current_statement_index)]
+            return [
+                (Token.In, 'In ['),
+                (Token.In.Number, '%s' % python_input.current_statement_index),
+                (Token.In, ']: '),
+            ]
 
         super(PythonPrompt, self).__init__(get_tokens)
 
 
 class PythonToolbar(TokenListToolbar):
-    def __init__(self, key_bindings_manager, settings, token=Token.Toolbar.Status):
+    def __init__(self, key_bindings_manager, python_input, token=Token.Toolbar.Status):
         def get_tokens(cli):
             python_buffer = cli.buffers['default']
 
@@ -150,7 +245,7 @@ class PythonToolbar(TokenListToolbar):
             append = result.append
 
             append((TB, ' '))
-            result.extend(get_inputmode_tokens(TB, key_bindings_manager, cli))
+            result.extend(get_inputmode_tokens(TB, key_bindings_manager, python_input, cli))
             append((TB, '  '))
 
             # Position in history.
@@ -158,15 +253,15 @@ class PythonToolbar(TokenListToolbar):
                                     len(python_buffer._working_lines))))
 
             # Shortcuts.
-            if not key_bindings_manager.enable_vi_mode and cli.focus_stack.current == 'search':
+            if not python_input.vi_mode and cli.focus_stack.current == 'search':
                 append((TB, '[Ctrl-G] Cancel search [Enter] Go to this position.'))
-            elif bool(cli.current_buffer.selection_state) and not key_bindings_manager.enable_vi_mode:
+            elif bool(cli.current_buffer.selection_state) and not python_input.vi_mode:
                 # Emacs cut/copy keys.
                 append((TB, '[Ctrl-W] Cut [Meta-W] Copy [Ctrl-Y] Paste [Ctrl-G] Cancel'))
             else:
                 append((TB, '  '))
 
-                if settings.paste_mode:
+                if python_input.paste_mode:
                     append((TB.On, '[F6] Paste mode (on)   '))
                 else:
                     append((TB.Off, '[F6] Paste mode (off)  '))
@@ -179,15 +274,16 @@ class PythonToolbar(TokenListToolbar):
         super(PythonToolbar, self).__init__(
             get_tokens,
             default_char=Char(token=token),
-            filter=~IsDone() & RendererHeightIsKnown())
+            filter=~IsDone() & RendererHeightIsKnown() &
+                Condition(lambda cli: python_input.show_status_bar and
+                                      not python_input.show_exit_confirmation))
 
 
-def get_inputmode_tokens(token, key_bindings_manager, cli):
+def get_inputmode_tokens(token, key_bindings_manager, python_input, cli):
     """
     Return current input mode as a list of (token, text) tuples for use in a
     toolbar.
 
-    :param vi_mode: (bool) True when vi mode is enabled.
     :param cli: `CommandLineInterface` instance.
     """
     mode = key_bindings_manager.vi_state.input_mode
@@ -197,7 +293,7 @@ def get_inputmode_tokens(token, key_bindings_manager, cli):
     append((token.InputMode, '[F4] '))
 
     # InputMode
-    if key_bindings_manager.enable_vi_mode:
+    if python_input.vi_mode:
         if bool(cli.current_buffer.selection_state):
             if cli.current_buffer.selection_state.type == SelectionType.LINES:
                 append((token.InputMode, 'Vi (VISUAL LINE)'))
@@ -222,12 +318,12 @@ def get_inputmode_tokens(token, key_bindings_manager, cli):
 
 
 class ShowSidebarButtonInfo(Window):
-    def __init__(self):
+    def __init__(self, python_input):
         token = Token.Toolbar.Status
 
         version = sys.version_info
         tokens = [
-            (token, ' [F2] Sidebar'),
+            (token, ' [F2] Options'),
             (token, ' - '),
             (token.PythonVersion, '%s %i.%i.%i' % (platform.python_implementation(),
                                                    version[0], version[1], version[2])),
@@ -241,12 +337,34 @@ class ShowSidebarButtonInfo(Window):
 
         super(ShowSidebarButtonInfo, self).__init__(
             TokenListControl(get_tokens, default_char=Char(token=token)),
-            filter=~IsDone() & RendererHeightIsKnown(),
+            filter=~IsDone() & RendererHeightIsKnown() &
+                Condition(lambda cli: python_input.show_status_bar and
+                                      not python_input.show_exit_confirmation),
             height=LayoutDimension.exact(1),
             width=LayoutDimension.exact(width))
 
 
-def create_layout(settings, key_bindings_manager,
+class ExitConfirmation(Window):
+    """
+    Display exit message.
+    """
+    def __init__(self, python_input, token=Token.ExitConfirmation):
+        def get_tokens(cli):
+            return [
+                (token, '\n Do you really want to exit? ([y]/n)'),
+                (Token.SetCursorPosition, ''),
+                (token, '  \n'),
+            ]
+
+        visible = ~IsDone() & Condition(lambda cli: python_input.show_exit_confirmation)
+
+        super(ExitConfirmation, self).__init__(
+            TokenListControl(get_tokens, default_char=Char(token=token),
+                             has_focus=visible),
+            filter=visible)
+
+
+def create_layout(python_input, key_bindings_manager,
                   python_prompt_control=None, lexer=PythonLexer, extra_sidebars=None,
                   extra_buffer_processors=None):
     D = LayoutDimension
@@ -261,8 +379,8 @@ def create_layout(settings, key_bindings_manager,
             """
             b = cli.buffers['default']
 
-            if b.complete_state is None and settings.signatures:
-                row, col = settings.signatures[0].bracket_start
+            if b.complete_state is None and python_input.signatures:
+                row, col = python_input.signatures[0].bracket_start
                 index = b.document.translate_row_col_to_index(row - 1, col)
                 return index
 
@@ -270,7 +388,7 @@ def create_layout(settings, key_bindings_manager,
             BufferControl(
                 buffer_name=DEFAULT_BUFFER,
                 lexer=lexer,
-                show_line_numbers=ShowLineNumbersFilter(settings, 'default'),
+                show_line_numbers=ShowLineNumbersFilter(python_input),
                 input_processors=[
                                   # Show matching parentheses, but only while editing.
                                   ConditionalProcessor(
@@ -304,37 +422,50 @@ def create_layout(settings, key_bindings_manager,
                         Float(xcursor=True,
                               ycursor=True,
                               content=CompletionsMenu(
+                                  scroll_offset=Integer.from_callable(
+                                      lambda: python_input.completion_menu_scroll_offset),
                                   max_height=12,
-                                  extra_filter=ShowCompletionsMenu(settings))),
+                                  extra_filter=show_completions_menu(python_input))),
                         Float(xcursor=True,
                               ycursor=True,
-                              content=SignatureToolbar(settings))
+                              content=MultiColumnCompletionsMenu(
+                                  extra_filter=show_multi_column_completions_menu(python_input))),
+                        Float(xcursor=True,
+                              ycursor=True,
+                              content=SignatureToolbar(python_input)),
+                        Float(left=2,
+                              bottom=1,
+                              content=ExitConfirmation(python_input)),
+                        Float(bottom=1, left=1, right=0, content=PythonSidebarHelp(python_input)),
                     ]),
                 ArgToolbar(),
                 SearchToolbar(),
                 SystemToolbar(),
                 ValidationToolbar(),
-                CompletionsToolbar(extra_filter=ShowCompletionsToolbar(settings)),
+                CompletionsToolbar(extra_filter=show_completions_toolbar(python_input)),
 
                 # Docstring region.
                 Window(height=D.exact(1),
                        content=FillControl('\u2500', token=Token.Separator),
-                       filter=HasSignature(settings) & ShowDocstring(settings) & ~IsDone()),
+                       filter=HasSignature(python_input) & ShowDocstring(python_input) & ~IsDone()),
                 Window(
                     BufferControl(
                         buffer_name='docstring',
                         default_token=Token.Docstring,
                         #lexer=PythonLexer,
                     ),
-                    filter=HasSignature(settings) & ShowDocstring(settings) & ~IsDone(),
+                    filter=HasSignature(python_input) & ShowDocstring(python_input) & ~IsDone(),
                     height=D(max=12),
                 ),
             ]),
             ] + extra_sidebars + [
-            PythonSidebar(settings, key_bindings_manager),
+            HSplit([
+                PythonSidebar(python_input),
+                PythonSidebarNavigation(python_input),
+            ])
         ]),
         VSplit([
-            PythonToolbar(key_bindings_manager, settings),
-            ShowSidebarButtonInfo(),
+            PythonToolbar(key_bindings_manager, python_input),
+            ShowSidebarButtonInfo(python_input),
         ])
     ])
